## todo: unarchive


# from utils_ak.interactive_imports import *
# from qset_feature_store import *
#
#
# class BasicStreamingDatasetManager:
#     def __init__(self, datasets_dic):
#         self.datasets_dic = {}
#         for k, v in datasets_dic.items():
#             df = qdm.get_dataset(v)
#             self.datasets_dic[k] = df
#
#         self.datasets_dic = {k: self.crop(v) for k, v in self.datasets_dic.items()}
#
#     def crop(self, dataset):
#         raise NotImplemented()
#
#     def get_dataset(self, featureset, params=None, **kwargs):
#         return self.datasets_dic[featureset['name']].copy()
#
#     def add(self, key, obj):
#         if isinstance(obj, dict):
#             new_df = pd.DataFrame(obj['values'], index=obj['index'], columns=obj['columns'])
#         elif isinstance(obj, pd.DataFrame):
#             new_df = obj
#         else:
#             raise Exception('Unknown row object type')
#
#         self.datasets_dic[key] = pd.concat([self.datasets_dic[key], new_df])  #todo: optimize, do not copy
#         self.datasets_dic[key] = self.crop(self.datasets_dic[key])
#
#
# class FixedLengthStreamingDatasetManager(BasicStreamingDatasetManager):
#     def __init__(self, datasets_dic, length):
#         self.length = length
#         super().__init__(datasets_dic)
#
#     def crop(self, dataset):
#         return dataset.iloc[-self.length:]
#
#
# class FixedPeriodStreamingDatasetManager(BasicStreamingDatasetManager):
#     def __init__(self, datasets_dic, period):
#         self.period = cast_timedelta(period) # inclusive!
#         super().__init__(datasets_dic)
#
#     def crop(self, dataset):
#         last_index = dataset.index[-1]
#         return dataset.loc[cast_datetime(last_index) - self.period:]
#
#
# class StreamingDatasetManager:
#     def __init__(self, featureset, session_gap=2):
#         self.featureset = featureset
#         self.session_gap = session_gap
#
#         config = cast_featureset_model(featureset).get_inherited_config()
#
#         datasets_dic = {k: v for k, v in config.get('params', {}).items() if isinstance(v, ObjectId) and cast_featureset_model(v).is_dataset()}
#
#         self.sdm = FixedPeriodStreamingDatasetManager(datasets_dic=datasets_dic,
#                                                       period=timedelta(seconds=config['materialization']['lookback_seconds']))
#
#         self.wm = BasicWindowManager(lambda key: ProcessingSessionWindow(gap=self.session_gap))
#
#         # todo: hardcode (local_file hook)
#         self.dataset_manager = DatasetManager({'local_file': self.sdm})
#
#     def process(self):
#         values = self.wm.close_if_possible()  # {ts: {dataset_name: dataset_value}}
#         if not values:
#             return
#
#         values = sum([v for v in values.values()], [])  # flatten
#         df = pd.DataFrame(values)  # ts, dataset, row
#         df = df.sort_values(by='ts')
#
#         for ts, grp_ts in df.groupby('ts'):
#             # add values for each dataset
#             for dataset, grp_dataset in grp_ts.groupby('dataset'):
#                 new_df = pd.DataFrame(grp_dataset['row'].values.tolist())
#                 new_df.index = [cast_datetime(ts)] * len(new_df)
#                 new_df.index.name = 'startRange'
#                 self.sdm.add(dataset, new_df)
#
#             print(self.dataset_manager.get_dataset(self.featureset))
#
#     def add(self, ts, value):
#         self.wm.add(key=ts, value=value)
#
#
# def test1():
#     from qset_feature_store import QsetRepositoryClient
#     cli = QsetRepositoryClient('akadaner')
#     sqdm = FixedPeriodStreamingDatasetManager({'timebars': cli.get_latest(user='akadaner', name='timebars')}, timedelta(minutes=1))
#     print(sqdm.datasets_dic)
#
#
# if __name__ == '__main__':
#     from qset_feature_store import QsetRepositoryClient
#     cli = QsetRepositoryClient('akadaner')
#     featureset = cli.get_latest(user='akadaner', name='feature_set1')
#     sqdm = StreamingDatasetManager(featureset, session_gap=2)
#
#     sqdm.add(cast_ts(datetime.datetime.now()), {'ts': str(datetime.datetime.now()), 'dataset': 'timebars', 'row': {'ticker': 'XTZ-USDT', 'pHigh': random.uniform(2.2, 2.4), 'pLow': random.uniform(2.0, 2.2)}})
#     time.sleep(3)
#     sqdm.add(cast_ts(datetime.datetime.now()), {'ts': str(datetime.datetime.now()), 'dataset': 'timebars', 'row': {'ticker': 'XTZ-USDT', 'pHigh': random.uniform(2.2, 2.4), 'pLow': random.uniform(2.0, 2.2)}})
#     sqdm.process()


import pandas as pd
from qset_feature_store.repository import models
from qset_feature_store.reflexion import get_featureset_code_from_current_notebook
from utils_ak.naming import to_lower_name
from utils_ak.builtin import delistify
from mongoengine import connect

from utils_ak.mongoengine import cast_model, cast_dict


class QsetRepositoryClient:
    def __init__(self, account, db='feature-store', host='localhost'):
        self.account = account
        connect(db, host=host)

    def _cast_mongoengine_query(self, query):
        query = query or {}
        if '_id' in query:
            query['pk'] = query.pop('_id')
        return query

    def get(self, type, **query):
        res = self.get_many(type, **query)
        return delistify(res)

    def get_many(self, type, **query):
        cls = models.cast_model_class(type)
        return [cast_dict(model, cls) for model in cls.objects(**self._cast_mongoengine_query(query))]

    def get_catalog(self, type, **query):
        return pd.DataFrame(self.get_many(type, **query))

    def get_latest(self, type, **query):
        cls = models.cast_model_class(type)

        latest = cls.objects(**self._cast_mongoengine_query(query)).last()

        if not latest:
            return

        return cast_dict(latest, cls)

    def delete(self, type, object):
        cls = models.cast_model_class(type)
        return cls.objects(pk=object['_id']).delete()

    def create_from(self, type, object):
        cls = models.cast_model_class(type)
        return cast_dict(cast_model(object, cls).clone(), cls)

    def publish(self, type, object):
        cls = models.cast_model_class(type)

        if '_id' in object:
            raise Exception('Feature Set already exists')
        res = cast_model(object, cls).save()
        return cast_dict(res, cls)

    def update(self, type, object):
        if '_id' not in object:
            raise Exception('Feature Set not published yet')
        new_object = self.create_from(type, object)
        return self.publish(type, new_object)

    def is_published(self, object):
        return '_id' in object

    def push(self, type, object):
        if not self.is_published(object):
            return self.publish(type, object)
        else:
            return self.update(type, object)

    def create_featureset(self, name, engine, body=None, inputs=None, **kwargs):
        body = body or {}
        inputs = inputs or {}
        return cast_dict(models.FeatureSet(account=self.account, name=name, engine=engine, body=body, inputs=inputs, **kwargs), models.FeatureSet)

    def create_pandas_featureset(self, featureset_cls, inputs=None, name=None, **kwargs):
        name = name or to_lower_name(featureset_cls.__name__)
        return self.create_featureset(name=name,
                                      engine='pandas/v1',
                                      inputs=inputs,
                                      body={'code': get_featureset_code_from_current_notebook(featureset_cls)}, **kwargs)

    def create_dataset_schema(self, name, featureset, params, **kwargs):
        featureset = cast_model(featureset, models.FeatureSet)

        return cast_dict(models.DatasetSchema(account=self.account,
                                              name=name,
                                              featureset=featureset,
                                              params=params,
                                              **kwargs), models.DatasetSchema)

    def create_dataset(self, name, scope, dataset_schema, materialization, **kwargs):
        dataset_schema = cast_model(dataset_schema, models.DatasetSchema)
        return cast_dict(models.DataSet(account=self.account,
                                        name=name,
                                        scope=scope,
                                        dataset_schema=dataset_schema,
                                        materialization=materialization, **kwargs), models.DataSet)


def test_qset_repository_client():
    cli = QsetRepositoryClient(account='akadaner', db='feature-store-test', host='localhost')
    fs = cli.create_featureset('my_featureset1',
                               engine='my_engine',
                               body={'code': 'print("Hello World")'},
                               metadata={'label': 'test'})
    fs = cli.push('featureset', fs)
    print(fs)
    print(cli.get_catalog('featureset'))
    dss = cli.create_dataset_schema('my_dataset_schema_1',
                                                featureset=fs,
                                                params={'param': 'foo'},
                                                metadata={'label': 'test'})
    dss = cli.push('dataset_schema', dss)
    print(dss)
    print(cli.get_catalog('dataset_schema'))
    ds = cli.create_dataset('my_dataset1',
                                        scope={'universe': 'my_universe'},
                                        dataset_schema=dss,
                                        materialization={'materialization_param': 'bar'},
                                        metadata={'label': 'test'})
    ds = cli.push('dataset', ds)
    print(ds)
    print(cli.get_catalog('dataset'))

    for key in ['featureset', 'dataset_schema', 'dataset']:
        for doc in cli.get_many(key, metadata={'label': 'test'}):
            print('Deleteting', key, doc['_id'])
            cli.delete(key, doc)
        print(cli.get_catalog(key))


if __name__ == '__main__':
    test_qset_repository_client()


from qset_feature_store.repository import QsetRepositoryClient
from qset_feature_store.featureset_engine.dataset_manager import LocalDatasetManager, EngineDatasetManager


class Controller:
    def __init__(self, account, repository_host='localhost'):
        self.account = account
        self.repository = QsetRepositoryClient(account=account, db='feature-store', host=repository_host)

        dms = []
        dms.append(LocalDatasetManager('data/'))
        self.datasets = EngineDatasetManager(dms, self.repository)



from utils_ak.time import cast_str
from datetime import datetime

from mongoengine import *
from utils_ak.mongoengine import cast_model, cast_dict


class UsersQuerySet(QuerySet):
    def last(self):
        return self.order_by('-id').first()

# todo: add branching


class BaseFeatureSet(Document):
    meta = {'abstract': True, 'queryset_class': UsersQuerySet}

    AUTO_FIELDS = ['created', 'version', '_id']
    name = StringField(default=lambda: cast_str(datetime.utcnow(), '%Y%m%d%H%M%S'))
    account = StringField(required=True)
    metadata = DictField()
    created = DateTimeField(default=datetime.utcnow)
    version = StringField(default=lambda: str(datetime.utcnow()))

    def clone(self):
        fs = cast_dict(self, self.__class__)
        for key in self.AUTO_FIELDS:
            fs.pop(key, None)
        return cast_model(fs, self.__class__)


class FeatureSet(BaseFeatureSet):
    engine = StringField(required=True)
    body = DictField()
    inputs = DictField()


class DatasetSchema(BaseFeatureSet):
    featureset = ReferenceField(FeatureSet, reverse_delete_rule=NULLIFY)
    params = DictField()


class DataSet(BaseFeatureSet):
    scope = DictField()
    dataset_schema = ReferenceField(DatasetSchema, reverse_delete_rule=NULLIFY)
    materialization = DictField()
    contents = DictField()
    state = DictField()


MODELS = {'FeatureSet': FeatureSet, 'featureset': FeatureSet,
          'DatasetSchema': DatasetSchema, 'dataset_schema': DatasetSchema,
          'DataSet': DataSet, 'dataset': DataSet}


def cast_model_class(model_class_name):
    return MODELS[model_class_name]




from qset_feature_store.featureset_engine.dataset_manager.dataset_manager import DatasetManager
from qset_feature_store.featureset_engine.dataset_manager.local_dataset_manager import LocalDatasetManager
from qset_feature_store.repository import models
from qset_feature_store.reflexion import extract_feature_set
from qset_feature_store.featureset_engine.featureset import FeatureSet
from utils_ak.builtin import list_get

# todo: pass repository client as an argument to engine dataset manager

class EngineDatasetManager(DatasetManager):
    def __init__(self, dataset_managers, repository_client):
        self.dataset_managers = dataset_managers
        self.ldm = list_get([dm for dm in self.dataset_managers if isinstance(dm, LocalDatasetManager)])
        assert self.ldm is not None
        self.repository_client = repository_client

    def get_from_dataset(self, dataset):
        raise Exception(f'Use get method to get dataset from dataset managers')

    def get_from_featureset(self, featureset, scope, params):
        # todo: make use of config properly
        featureset = models.cast_dict(featureset, models.FeatureSet)
        assert featureset['engine'] == 'pandas/v1'
        cls = extract_feature_set(featureset['body']['code'])
        featureset_instance = cls()
        return self.calc(featureset_instance, featureset['inputs'], scope, params)

    def calc(self, featureset_instance, inputs, scope, params):
        featureset_instance.inputs = inputs
        featureset_instance.params = params
        featureset_instance.scope = scope
        featureset_instance.dataset_provider = self
        return featureset_instance.calc()

    def get_from_dataset_schema(self, dataset_schema, scope):
        # fetch local dataset if possible
        datasets = self.repository_client.get_many('dataset', dataset_schema=dataset_schema['_id'], scope=scope, materialization={'storage': 'local'})

        if datasets:
            assert len(datasets) == 1, 'Only one local dataset allowed for current dataset schema'
            try:
                return self.get(datasets[0])
            except:
                pass

        featureset = models.cast_dict(dataset_schema['featureset'], models.FeatureSet)
        return self.get_from_featureset(featureset, scope, dataset_schema['params'])

    def get(self, generic_featureset, scope=None, params=None):
        generic_featureset = super()._format_generic_featureset(generic_featureset)

        # try to get from data managers first
        for manager in self.dataset_managers:
            try:
                return manager.get(generic_featureset, scope, params)
            except:
                pass

        # try to calc otherwise
        return super().get(generic_featureset, scope, params)

    def put(self, df, dataset):
        assert dataset.get('materialization', {}).get('storage') == 'local', 'Only local storage is supported.'
        return self.ldm.put(df, dataset)



def test_calc():
    import sys
    sys.path.append(r'C:\Users\Mi\Desktop\master\code\git\2020.09-qset\qset-feature-store')
    from qset_feature_store import FeatureSet
    import pandas as pd
    class TrivialFeatureSet(FeatureSet):
        def calc(self):
            return pd.DataFrame([1, 2, 3], columns=['a'])

    import sys
    sys.path.append(r'C:\Users\Mi\Desktop\master\code\git\2020.09-qset\qset-feature-store')
    from qset_feature_store import FeatureSet
    import pandas as pd
    class FeatureSetUsingTimebars(FeatureSet):
        def calc(self):
            return self.get(self.inputs['timebars']) * 2

    from qset_feature_store.repository.client import QsetRepositoryClient
    cli = QsetRepositoryClient('akadaner', db='feature-store-test')

    dms = []
    dms.append(LocalDatasetManager('data/'))
    qdm = qset_dataset_manager = EngineDatasetManager(dms, cli)

    # clean
    for key in ['featureset', 'dataset_schema', 'dataset']:
        for doc in cli.get_many(key):
            cli.delete(key, doc)

    # create timebars featureset
    fs = cli.create_featureset('timebars', 'external')
    fs = cli.push('featureset', fs)

    dss = cli.create_dataset_schema('timebars_schema', fs, {'period': 600})
    dss = cli.push('dataset_schema', dss)

    ds = cli.create_dataset('timebars_local', scope={'universe': 'binance'}, dataset_schema=dss, materialization={'storage': 'local'})
    ds = cli.push('dataset', ds)

    df = pd.DataFrame([[1, 4], [2, 5], [3, 6]], columns=['a', 'b'])

    qdm.put(df, ds)

    print(qdm.get(ds).head())
    print(qdm.calc(TrivialFeatureSet(), inputs={}, scope={'universe': 'binance'}, params={}))
    print(qdm.calc(FeatureSetUsingTimebars(), inputs={'timebars': dss}, scope={'universe': 'binance'}, params={}))

    # create configured with pandas

    code = r'''
import sys
sys.path.append(r'C:\Users\Mi\Desktop\master\code\git\2020.09-qset\qset-feature-store')
from qset_feature_store import FeatureSet
import pandas as pd
class TrivialFeatureSet(FeatureSet):
    def calc(self):
        return pd.DataFrame([1, 2, 3], columns=['a'])
'''

    fs = cli.create_featureset('my_pandas_featureset', 'pandas/v1', {'code': code})
    fs = cli.push('featureset', fs)
    print(fs)

    # note: using get now - since it is a repository featureset
    print(qdm.get(fs, scope={'universe': 'binance'}, params={}))


if __name__ == '__main__':
    test_calc()


from qset_feature_store.repository import models

# todo: refactor code duplicate


class DatasetManager:
    def _format_generic_featureset(self, generic_featureset):
        if generic_featureset['_cls'] == 'DataSet':
            return models.cast_dict(generic_featureset, models.DataSet)
        elif generic_featureset['_cls'] == 'FeatureSet':
            return models.cast_dict(generic_featureset, models.FeatureSet)
        elif generic_featureset['_cls'] == 'DatasetSchema':
            return models.cast_dict(generic_featureset, models.DatasetSchema)

    def get(self, generic_featureset, config=None, params=None):
        generic_featureset = self._format_generic_featureset(generic_featureset)
        if generic_featureset['_cls'] == 'DataSet':
            return self.get_from_dataset(generic_featureset)
        elif generic_featureset['_cls'] == 'FeatureSet':
            return self.get_from_featureset(generic_featureset, config, params)
        elif generic_featureset['_cls'] == 'DatasetSchema':
            assert params is None
            return self.get_from_dataset_schema(generic_featureset, config)

    def put(self, df, dataset):
        assert dataset['_cls'] == 'DataSet' and dataset.get('materialization', {}).get('storage') == 'local'
        generic_featureset = self._format_generic_featureset(dataset)
        return self.put_dataset(df, generic_featureset)

    def get_from_dataset(self, dataset):
        raise NotImplementedError

    def get_from_featureset(self, featureset, config, params):
        raise NotImplementedError

    def get_from_dataset_schema(self, dataset_schema, config):
        raise NotImplementedError

    def put_dataset(self, df, dataset, **kwargs):
        raise NotImplementedError


import asyncio
from utils_ak.mongo_job_queue.worker.worker import Worker
from utils_ak.mongo_job_queue.worker.worker_test import *

# from qset_feature_store import qdm, QsetRepositoryClient


class DatasetWorker(Worker):
    def __init__(self, *args, **kwargs):
        # todo: use secrets
        super().__init__(*args, **kwargs)
        # self.cli = QsetRepositoryClient('akadaner', host='mongodb+srv://arseniikadaner:Nash0lsapog@cluster0.2umoy.mongodb.net/feature-store?retryWrites=true&w=majority')

    async def process(self):
        # fs = self.cli.get(id=self.payload['featureset_id'])

        # self.microservice.logger.info('Calculating featureset', featureset_id=fs['_id'])
        self.microservice.logger.info('Calculating featureset', featureset_id=self.payload['featureset_id'])

        # todo: switch-com
        # df = qdm.get_dataset(fs)
        time.sleep(5)

        # todo: save dataset to output storage (clickhouse)

        # todo: switch-com
        # self.microservice.send_state('success', {'calculated_length': len(df)})

        # self.microservice.logger.info('Successfully calculated featureset', featureset_id=fs['_id'])
        self.microservice.logger.info('Successfully calculated featureset', featureset_id=self.payload['featureset_id'])
        self.microservice.send_state('success', {'calculated_length': 78078})


def test_batch():
    test_worker(DatasetWorker, {'featureset_id': '602020f1079228082591d0cf'}, message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))


def test_streaming():
    test_worker(DatasetWorker, {'featureset_id': '602020f1079228082591d0cf'}, message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))


def test_deployment():
    test_worker_deployment('deployment.yml', message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))


if __name__ == '__main__':
    # test_batch()
    # test_streaming()
    test_deployment()


from utils_ak.interactive_imports import *
import qset_feature_store

import re


def get_featureset_code_from_current_notebook(featureset):
    code = get_current_notebook_code()
    cells = parse_cells(code)

    featureset_class_name = re.search(r"'(.+)'", str(featureset)).groups()[0].split('.')[-1]
    cell = [cell for cell in cells if f'class {featureset_class_name}(' in cell][0]
    cell = re.sub(r'\n+', '\n', cell)
    cell = re.sub(r'^\n', '', cell)
    cell = re.sub(r'\n$', '', cell)
    return cell


def _is_featureset_child(obj):
    if not inspect.isclass(obj):
        return False

    if obj == qset_feature_store.FeatureSet:
        return False

    if not any(['FeatureSet' in str(base) for base in obj.__mro__]):
        return False

    try:
        return isinstance(obj(), qset_feature_store.FeatureSet)
    except:
        return False


def extract_feature_set(code):
    with tempfile(suffix='.py') as fn:

        with open(fn, 'w') as f:
            f.write(code)

        classes = extract_all_classes(fn)
        classes = {k: v for k, v in classes.items() if _is_featureset_child(classes[k])}
        assert len(classes) == 1
        key = list(classes.keys())[0]
        return classes[key]



from utils_ak.loguru import *

from qset_feature_store import *
from qset_feature_store.dataset_manager.dataset_manager import qdm
from qset_feature_store.repository import QsetRepositoryClient


CLI = QsetRepositoryClient('akadaner')


class HistoricalServer(ProductionMicroservice):
    def __init__(self):
        super().__init__()
        self.add_callback('datasets', 'historical', self.on_dataset)

    def on_dataset(self, topic, msg):
        featureset = CLI.get(_id=cast_object_id(msg['featureset']['_id']))
        print(qdm.get_dataset(featureset))


class Ping(ProductionMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__('Ping microservice', *args, **kwargs)

    def send_ping(self, topic, msg):
        self.logger.info(f'Received {topic} {msg}')
        time.sleep(3)


def run_ping():
    ping = Ping()

    async def send_initial():
        await asyncio.sleep(3.0)
        try:
            ping.publish_json('datasets', 'historical', {'featureset': {'_id': str(CLI.get_latest(user='akadaner', name='timebars')['_id'])}})
        except Exception as e:
            import traceback
            traceback.print_exc()
            print('hey')
        # ping.publish_json('datasets', 'historical', {})
        print('sent')

    ping.tasks.append(asyncio.ensure_future(send_initial()))
    ping.run()


if __name__ == '__main__':
    multiprocessing.Process(target=run_ping).start()
    server = HistoricalServer()
    # run_listener_async('heartbeat', brokers_config=server.brokers_config)
    server.run()



class FeatureSet:
    def __init__(self):
        self.params = {}
        self.inputs = {}

        self.scope = None
        self.dataset_provider = None

    def calc(self):
        raise Exception('Not implemented')

    def get(self, generic_featureset, params=None):
        if self.dataset_provider is None:
            raise Exception('Data provider not set')
        return self.dataset_provider.get(generic_featureset, scope=self.scope, params=params)
